import os
import shutil
import subprocess

from tokit.utils import make_rand


def create_project(name):
    print("This will create a project skeleton using Tokit in current directory. GNU sed is requied.")
    source = os.path.join(os.path.dirname(__file__), 'skeleton')
    destination = os.path.join(os.getcwd(), name)
    print("Destination: ", destination)
    shutil.copytree(source, destination)

    commands = [
        "cd " + destination,
        """LC_ALL=C find config -type f \
            -exec sed -i s/PROJECT/%s/g {} \; """ % name,
        "python3 -m venv .",
        "bin/python3 -m ensurepip",
        "pip3 install -r requirements.txt",
        "bin/python3 src/app.py"
    ]
    subprocess.call(' && '.join(commands), stderr=subprocess.STDOUT, shell=True)

if __name__ == '__main__':
    name = input("Enter project name (empty to quit): ")
    if name:
        create_project(name)



import binascii
import hashlib
import json
from contextlib import contextmanager
from datetime import datetime
from json import JSONEncoder
from uuid import UUID
import shortuuid


class Event:
    """
    Event handlers storage.

    Example:

    >>> def handler(**kwargs):
    ...     print("Triggered:", kwargs)
    ...
    >>> Event.get('some_thing_happened').attach(handler)
    >>> Event.get('some_thing_happened').emit(status='OK')
    Triggered: {'status': 'OK'}

    """

    _repo = {}

    def __init__(self, name):
        self.handlers = []
        self.name = name

    def attach(self, handler, priority=0):
        handler._event_priority = priority
        self.handlers.append(handler)
        self.handlers.sort(key=lambda h: h._event_priority)

    def detach(self, handler):
        self.handlers.remove(handler)

    @classmethod
    def get(cls, name):
        instance = cls._repo.get(name, None)
        if not instance:
            instance = cls(name)
            cls._repo[name] = instance
        return instance

    @contextmanager
    def subscribe(self, *handlers):
        for handler in handlers:
            self.attach(handlers)
        try:
            yield
        finally:
            for handlers in handlers:
                self.detach(handlers)

    def emit(self, *args, **kwargs):
        for handler in self.handlers:
            handler(*args, **kwargs)


def on(event_name, priority=0):
    def decorator(fn):
        Event.get(event_name).attach(fn, priority)
    return decorator


class AttributeDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__

class VersatileEncoder(JSONEncoder):
    """
    Encode all "difficult" object such as UUID
    """

    def _iterencode(self, obj, markers=None):

        if isinstance(obj, tuple) and hasattr(obj, '_asdict'):
            gen = self._iterencode_dict(obj._asdict(), markers)
        else:
            gen = JSONEncoder._iterencode(self, obj, markers)
        for chunk in gen:
            yield chunk

    def default(self, obj):
        if isinstance(obj, UUID):
            return shortuuid.encode(obj)
        elif isinstance(obj, datetime):
            return str(obj)
        else:
            return JSONEncoder.default(self, obj)

def to_json(obj):
    return json.dumps(obj, ensure_ascii=False, cls=VersatileEncoder).replace("</", "<\\/")

def make_rand(length=16):
    shortuuid.ShortUUID().random(length)

def make_hash(secret, **kwargs):
    kwargs = kwargs or dict(
        salt=b'tokit',
        iterations=82016,
    )
    # might be GIL dependent
    dk = hashlib.pbkdf2_hmac('sha512', str.encode(secret), **kwargs)
    return binascii.hexlify(dk).decode()

try:
    import markdown


    def md_html(raw):
        """
        Convert Markdown format string to HTML
        WARNING: Be careful of leading spaces
        """
        html = markdown.markdown(raw.decode())
        return html
except:
    pass


"""
Transform to browsers' languages
Can serve files directly using shortcut: `python3 -m tokit.compiler`

Require system's libsass and pyv8
"""
import os
import io
import tornado.ioloop
import tornado.web
from tornado.iostream import IOStream

COMPILER_URLS = []

def read_file(filename):
    with io.open(filename, encoding='utf8') as fp:
        return fp.read()


class CompilerHandler(tornado.web.RequestHandler):

    def set_default_headers(self):
        self.set_header('Server', 'Python3')

    def get(self, requested_file):
        try:
            requested_path = requested_file.replace(self.application.settings['static_url_prefix'], '/')
            full_path = self.application.root_path + requested_path
            # TODO check harmful path
            self.compile(full_path)
        except Exception as e:
            self.set_status(400)
            self.write('/*')
            self.write(str(e))
            self.write('*/')

try:
    import sass

    class SassHandler(CompilerHandler):

        def prepare(self):
            self.set_header('Content-Type', 'text/css')

        def compile(self, full_path):

            self.write(sass.compile(
                filename=full_path,
                output_style=('nested' if self.application.settings['debug'] else 'compressed')
            ))

    COMPILER_URLS.append((r'^(/.+\.sass)$', SassHandler))
except ImportError:
    pass

try:
    import execjs

    class CoffeeHandler(CompilerHandler):

        def prepare(self):
            self.set_header('Content-Type', 'application/javascript')
            self.context = execjs.get().compile(
                read_file(os.path.dirname(__file__) + '/js/coffee-script.js')
            )

        def compile(self, full_path):
            result = self.context.call(
                "CoffeeScript.compile",
                read_file(full_path),
                {'bare': True}
            )
            self.write(result)

    class RiotHandler(CompilerHandler):

        def prepare(self):
            self.set_header('Content-Type', 'application/javascript')
            self.context = execjs.get().compile(
                read_file(os.path.dirname(__file__) + '/js/coffee-script.js') +
                ";\n\n var exports = {}; module.exports = {};" +  # fake CommonJS environment
                read_file(os.path.dirname(__file__) + '/js/riotc.js') + "; var riot = module.exports;"
            )

        def compile(self, requested_file):
            result = self.context.call(
                "riot.compile",
                read_file(self.application.root_path + requested_file),
                True
            )
            self.write(result)


    COMPILER_URLS.append((r'^(/.+\.coffee)$', CoffeeHandler))
    COMPILER_URLS.append((r'^(/.+\.tag)$', RiotHandler))
except ImportError:
    pass

def init_complier(app):
    from tokit import logger
    if len(COMPILER_URLS):
        app.add_handlers('.*$', COMPILER_URLS)
        app.root_path = app.config.root_path
    else:
        logger.warn('Found no compilers handlers')

if __name__ == "__main__":
    app = tornado.web.Application(urlspecs())
    app.root_path = os.path.realpath(os.path.curdir)
    print("Serving via compiler in:", app.root_path)
    app.listen(7998)
    tornado.ioloop.IOLoop.current().start()



import logging
import os
from uuid import UUID

import shortuuid
import cassandra
from cassandra.cluster import Cluster
from cassandra.cqlengine import connection as cqlengine_connection
from cassandra.query import dict_factory
from tornado.concurrent import Future
from tornado.ioloop import IOLoop

import tokit

logger = logging.getLogger(__name__)


def allow_management():
    os.environ['CQLENG_ALLOW_SCHEMA_MANAGEMENT'] = 'CQLENG_ALLOW_SCHEMA_MANAGEMENT'


def cassandra_init(app):
    """
    Sample configuration::

        [cassandra]
        contact_points=
            host1
            host2
        port=9042
        keyspace=blabla
    """
    logging.getLogger('cs').setLevel(tokit.logger.getEffectiveLevel())
    try:
        config = app.config.env['cassandra']
    except KeyError:
        logger.warn('Cassandra was not configured')
        return
    hosts = [p.strip() for p in config.get('contact_points').strip().split('\n')]
    logger.debug('%s', hosts)

    # for use with controllers
    cluster = Cluster(
        contact_points=hosts,
        port=int(config.get('port')),
        connect_timeout=1,
    )
    keyspace = app.config.env['cassandra'].get('keyspace')
    session = cluster.connect(keyspace)
    session.row_factory = dict_factory
    app.cs_pool = session
    
    # for use with object mapper
    cqlengine_connection.setup(
        hosts,
        default_keyspace=config.get('keyspace'),
        lazy_connect=True, retry_connect=True
    )
    # Further hook
    tokit.Event.get('cassandra_init').emit(app)


tokit.Event.get('init').attach(cassandra_init)

def serialize(row):
    if 'id' in row:
        row['short_id'] = shortuuid.encode(row['id'])
    return row


class IntegrityError(BaseException):
    """ this never happen because row will be overrided """
    pass


class CassandraMixin:
    """
    Reference: http://datastax.github.io/python-driver/api/cassandra/cluster.html#cassandra.cluster.Session.execute_async
    """

    DbIntegrityError = IntegrityError
    DbError = cassandra.RequestExecutionException

    @property
    def cs_pool(self):
        return self.application.cs_pool

    def cs_query(self, cql, *args):
        future = Future()
        cs_future = self.cs_pool.execute_async(cql, args)

        def _success(result):
            IOLoop.instance().add_callback(future.set_result, result)

        def _fail(e):
            IOLoop.instance().add_callback(future.set_exception, e)

        cs_future.add_callbacks(_success, _fail)
        return future

    async def cs_one(self, table, row_id):
        result = await self.cs_query(
            "SELECT * FROM " + table + " WHERE id = %s ",
            row_id
        )
        if result:
            return serialize(result[0])

    async def cs_upsert(self, table, **data):
        id = data.get('id', UUID())
        cql = "UPDATE {table} SET ".format(table=table) + ", ".join([
            k + " = %s " for k in data.keys()
        ]) + "WHERE id = %s"
        result = await self.cs_query(cql, list(data.values()) + [data['id']])
        return result

    async def cs_insert(self, table, **data):
        fields = data.keys()
        cql = 'INSERT INTO {} ({}) VALUES ({})'.format(table,
            ','.join(fields),
            ','.join(['%s'] * len(fields))
        )
        result = await self.cs_query(cql, *list(data.values()))
        return result

    db_insert = cs_upsert
    db_update = cs_upsert
    db_one = cs_one



import logging
import shortuuid
import uuid

import momoko
from psycopg2.extras import DictCursor
from psycopg2.extras import DictRow
import psycopg2.extensions

from tornado.gen import coroutine
from sqlbuilder.smartsql import Table, Query

import tokit
from tokit import api

logger = tokit.logger

class DictLogCursor(DictCursor):

    def execute(self, sql, args=None):
        logger.debug('Excute SQL: %s', self.mogrify(sql, args).decode())
        return super().execute(sql, args)


@tokit.on('init')
def pg_init(app):
    """ Hook to init Postgres momoko driver.
    dsn config is required, with syntax same as Psycopg2 DSN.

    Sample env.ini::

        [postgres]
        dsn=dbname=[APP_NAME]
        size=2
    """
    logging.getLogger('momoko').setLevel(logger.getEffectiveLevel())
    env = app.config.env['postgres']
    momoko_opts = dict(
        dsn=env['dsn'],
        size=int(env['size']),
        max_size=int(env['max_size']),
        auto_shrink=env.getboolean('auto_shrink'),
        cursor_factory=(DictLogCursor if env.getboolean('log') else DictCursor),
        # connection_factory=env.get('connection_factory', None),
    )
    app.pg_db = momoko.Pool(**momoko_opts)
    app.pg_db.connect()


class PgMixin:

    DbIntegrityError = psycopg2.IntegrityError
    DbError = psycopg2.Error

    @property
    def db(self):
        return self.application.pg_db

    @coroutine
    def pg_insert(self, table, fields=None, **data):
        """
        Postgres shorcut to insert data
        :return int new row's id

        Example::

            user_id = yield self.pg_insert('users', {"username": "foo", "password": "secret"})
        """
        if fields:
            data = self.get_request_dict(*fields)
        else:
            fields = list(data.keys())
        assert len(data) > 0  # check data
        values = list(data.values())

        sql = 'INSERT INTO {} ({}) VALUES ({}) RETURNING id ' \
            .format(table,
                    ','.join(fields),
                    ','.join(['%s'] * len(fields))
                    )
        cursor = yield self.db.execute(sql, values)
        return cursor.fetchone()[0]

    @coroutine
    def pg_update(self, table, data):
        changes = [field + ' = %s' for field in data.keys()]
        sql = 'UPDATE {} SET {} WHERE id = %s'.format(table, ','.join(changes))
        values = list(data.values()) + [data['id']]
        yield self.db.execute(sql, values)

    @coroutine
    def pg_query(self, query, *params):
        """ Low level execuation """
        result = yield self.db.execute(query, params)
        return result

    def pg_serialize(self, row):
        if not row:
            return
        ret = dict(row) if isinstance(row, DictRow) else row
        return ret

    @coroutine
    def pg_select(self, query, *params):
        """ Query and convert result """
        result = yield self.pg_query(query, *params)
        return [self.pg_serialize(row) for row in result.fetchall()]

    @coroutine
    def pg_one(self, query, *params):
        result = yield self.pg_query(query, *params)
        row = result.fetchone()
        if row:
            return self.pg_serialize(row)

    db_insert = pg_insert
    db_update = pg_update
    db_query = pg_query
    db_select = pg_select
    db_one = pg_one


class UidMixin:

    def pg_serialize(self, row):
        ret = PgMixin.pg_serialize(self, row)
        if 'id' in ret:
            ret['short_id'] = shortuuid.encode(uuid.UUID(ret['id']))
        return ret



import re
import os
from tornado.template import Loader, Template

SHORTCUT_RE = [
    (re.compile(rb'{\:\:(.*?)}', re.DOTALL),    rb'{{ _("\1") }}'),
    (re.compile(rb'{\:(.*?)}', re.DOTALL),      rb'{{ _(\1) }}'),
    (re.compile(rb'{=(.*?)}', re.DOTALL),       rb'{{ \1 }}'),
]


class CustomLoader(Loader):
    """
    Tornado's template preprocessor with translation shortcut.
    Key must be in single line

        * ``{::key}`` -> ``{{ _("key") }}``
        * ``{:key}``-> ``{{ _(key) }}``
        * ``{=expr}``-> ``{{ expr }}``
    """

    def _custom_prepocessor(self, content):
        ret = content
        for regex, replacement in SHORTCUT_RE:
            ret = regex.sub(replacement, ret)
        return ret

    def _create_template(self, name):
        path = os.path.join(self.root, name)
        with open(path, "rb") as f:
            template = Template(
                self._custom_prepocessor(f.read()),
                name=name, loader=self,
                compress_whitespace=False)
            return template


class TranslationMixin:

    def create_template_loader(self, template_path):
        settings = self.application.settings
        if "template_loader" in settings:
            return settings["template_loader"]
        kwargs = {}
        if "autoescape" in settings:
            # autoescape=None means "no escaping", so we have to be sure
            # to only pass this kwarg if the user asked for it.
            kwargs["autoescape"] = settings["autoescape"]
        return CustomLoader(template_path, **kwargs)



import os
import random
import string

from fabric3.operations import local, put
from fabric3.context_managers import lcd, cd, hide
from fabric3.contrib.project import rsync_project
from fabric3.contrib.files import exists
from fabric3.decorators import runs_once
from fabric3.api import run, env, hosts
from fabric3.api import settings
from fabric3.colors import green, red
from tokit.utils import rand

# Skip using git, imply force deploy
SKIP_PREPARE = os.environ.get('SKIP_PREPARE', False)

# Force deploy: override existing deployments
FORCE_DEPLOY = os.environ.get('FORCE_DEPLOY', SKIP_PREPARE)

DEPLOY_BRANCH = os.environ.get('DEPLOY_BRANCH', 'origin/master')
DEPENDENT_FILE = 'requirements.txt'
RSYNC_EXCLUDES = [
    '/.git*', '/tmp', '.env', '/logs'
]
LINK_DIRS = []
LINK_FILES = []

current_dir = os.path.abspath(os.path.dirname(__file__))

stamp = rand(6)


# if Git is not clean, save stash and store current branch
if not SKIP_PREPARE:
    with hide('output', 'running'):
        dirty_branch = None
        is_dirty = local('git diff --shortstat 2> /dev/null | tail -n1', capture=True)
        if is_dirty:
            print(red("Having ongoing change code, will save it"))
            dirty_branch = local('git rev-parse --abbrev-ref HEAD', capture=True)

def get_version():
    if FORCE_DEPLOY:
        # force deploy by create a special version
        v = 'force_' + stamp
    else:
        v = local('git rev-parse --short %s' % DEPLOY_BRANCH, capture=True)
    return v

def remove_old_versions(path):
    keep_last = 5
    with cd(path):
        # remove old old
        run("ls -1tr | head -n -%d | xargs -d '\n' rm -Rf " % keep_last)

def check(dest_path):
    test_domain = os.environ.get('TEST_DOMAIN')
    if not test_domain:
        return

    with cd(dest_path):
        # query a test URL using new code to see if it works
        run('ln -sfn versions/%s test' % get_version())
        status = run(
            'curl -sL -w "%{http_code} %{url_effective}\\n" ' +
            '--header "Host: ' + test_domain + '" -o /dev/null ' +
            'http://127.0.0.1:80/'
        )
        if '200' not in status:
            raise RuntimeError('Domain test failed')

def deploy_to(dest_path):
    v = get_version()

    if not exists(dest_path + '/versions'):
        run('mkdir -p %s/versions' % dest_path)

    if not exists(dest_path + '/versions/' + v):
        rsync_project(
            dest_path + '/versions/tmp', current_dir + '/',
            exclude=RSYNC_EXCLUDES,
            extra_opts='--delete-after --links'
        )
        # links
        with cd(dest_path):
            run('cp -R versions/tmp versions/%s' % v)

    with cd(dest_path + '/versions/' + v):
        for d in LINK_DIRS:
            shared_d = '%s/shared/%s' % (dest_path, d)
            if not exists(shared_d):
                run('mkdir -p ' + shared_d)
                run('chmod 770 ' + shared_d)
            run('ln -sfn %s .' % (shared_d))

        for f in LINK_FILES:
            run("ln -sf {dest}/shared/{f} {f}".format(
                f=f, dest=dest_path
            ))


def finalize(dest_path):
    v = get_version()
    with cd(dest_path):
        run('ln -sfn versions/%s current' % v)
    remove_old_versions(dest_path + '/versions/')

def restore():
    if SKIP_PREPARE:
        return
    if dirty_branch:
        local('git checkout ' + dirty_branch)
        local('git stash apply')
    else:
        local('git checkout master')

def prepare(dest_path):
    if SKIP_PREPARE:
        return
    if dirty_branch:
        local('git stash')
    local('git fetch --all', capture=True)
    local('git checkout %s' % DEPLOY_BRANCH, capture=True)

    update_dependents = True
    if exists(dest_path + '/current'):
        current_path = run('readlink -f %s/current' % dest_path)
        current_hash = os.path.basename(current_path.strip())
        if current_hash == get_version():
            raise RuntimeError('Already deployed version ' + current_hash)
        if 'force_' not in current_hash:
            changes = local(
                'git diff --name-only origin/master ' + current_hash,
                capture=True
            )
            if not DEPENDENT_FILE in changes:
                update_dependents = False

    if update_dependents:
        run('pip install -r requirements.txt')

def deploy(srv_path):
    try:
        with hide('output', 'running'):
            prepare(srv_path)
            deploy_to(srv_path)
            check(srv_path)
            finalize(srv_path)
        print(green("Success"))
    except RuntimeError as e:
        print(red(e.message))
        restore()


@runs_once
def ask_version(ver, dest):
    with hide('output', 'running'):
        if not ver:
            current_path = run('readlink -f %s/current' % dest)
            current_hash = os.path.basename(current_path.strip())
            print("Current version", green(current_hash))
            versions = run('ls -t {dest}/versions'.format(dest=dest))
            versions = re.split('\s+', versions.replace('tmp', '').replace(current_path, ''))
            # first is often current version
            ver = versions[1]
        modified = run('stat --format="%y" {dest}/versions/{ver}'.format(dest=dest, ver=ver))
    from fabric.contrib.console import confirm
    print(red("[IMPORTANT]"))
    print("Detected last version: " + yellow(ver))
    print("All versions: ", ', '.join(versions))
    ok = confirm(yellow(
        "Do you want to revert to  " + ver + " version? Deployed at " + modified[0:16],
        default=False
    ))
    return ver if ok else False

def revert(dest_path, ver=None):
    ver = ask_version(ver, dest_path)
    if ver:
        run((
            'cd {dest_path}; '
            'if test -d versions/{v} ; then '
                'ln -sfn versions/{v} current ; '
            'fi'
        ).format(v=ver, dest_path=dest_path))
        print(green("Reverted into" + ver))
    else:
        print(red("Canceled"))


class _X:
    """ Custom variables """
    app = None
    base_port = 7000
    n_instances = 2
    remote_path = '/home/'
    wait = 3
    excludes = ('.*', 'tmp', '__pycache__', '*.pyc', 'upload', 'cache')
    dirs = ['config', 'src', 'doc']


env.x = _X()

def pwd(base_file):
    _pwd = os.path.dirname(base_file)
    os.chdir(_pwd)

def with_root():
    env.user = 'root'

@contextlib.contextmanager
def on_remote(path=''):
    with cd(os.path.join(env.x.remote_path, path)):
        yield


@contextlib.contextmanager
def venv(inside=''):
    with prefix('source {remote}/bin/activate' \
                        .format(remote=env.x.remote_path)), cd(env.x.remote_path + inside):
        yield

def sync(dirs=None):
    if dirs:
        dirs = dirs.split(',')
    else:
        dirs = env.x.dirs
    for d in dirs:
        fab_project.rsync_project(
            remote_dir=env.x.remote_path, local_dir=d,
            exclude=env.x.excludes,
            extra_opts='--delete-after'
        )


def backend(action='restart'):
    """ (3) Start services """
    with_root()
    systemd_reload()
    for instance in range(env.x.n_instances):
        run('systemctl {action} {app}@{port}.service'.format(
            action=action,
            app=env.x.app,
            port=env.x.base_port + instance))
        time.sleep(env.x.wait)


def config():
    """ (3) Link configs """
    with_root()
    with on_remote():
        for instance in range(env.x.n_instances):
            service_file = "/etc/systemd/system/multi-user.target.wants/{app}@{port}.service" \
                .format(app=env.x.app, port=env.x.base_port + instance)
            run("ln -s -f {path}/config/app@.service {dest}" \
                .format(path=env.x.remote_path, dest=service_file))
        run('ln -s -f {path}/config/nginx.conf /etc/nginx/conf.d/{app}.conf' \
            .format(app=env.x.app, path=env.x.remote_path))


def rollback():
    with on_remote():
        run('mv src src-fail && mv rollback src')
        backend('restart')


def ssh_setup(public_key='~/.ssh/id_rsa.pub'):
    """ (0) Create user, add SSH key, create folders """
    env.user = 'root'
    run('useradd %s' % env.x.app)
    run('mkdir -p /home/%s/{.ssh,tmp,shared,src,var}' % env.x.app)
    run('ssh-keygen -b 2048 -t rsa -f /home/%s/.ssh/id_rsa -q -N ""' % env.x.app)
    put(public_key, '/home/%s/.ssh/authorized_keys' % env.x.app)
    run('chown %s -R /home/%s' % (env.x.app, env.x.app))
    run('chmod 700 /home/%s' % env.x.app)


def authorize_key(key=None):
    """ Add an user to access server """
    if key:
        run('cat "%s">> /home/%s/.ssh/authorized_keys' % (key, env.x.app))
    else:
        print("Please provide a SSH public key")


def static_copy():
    with_root()
    run('mkdir -p /var/www/{app}'.format(app=env.x.app))
    run('cp -R {path}/src/static /var/www/{app}/'.format(path=env.x.remote_path, app=env.x.app))
    run('chown -R nginx /var/www/{app}/'.format(path=env.x.remote_path, app=env.x.app))


def systemd_reload():
    with_root()
    run('systemctl daemon-reload')


def up_file(name):
    os.path.dirname(name)
    rel = os.path.relpath(name, PWD)
    print(put(name, os.path.join(env.x.remote_path, rel)))


def up_python():
    with_root()
    backend('restart')


def up_nginx():
    # Pitfall
    with_root()
    config()
    systemd_reload()
    run('nginx -t && systemctl reload nginx.service')


def nginx(action='restart'):
    with_root()
    # PITFALL: first start must be a restart, not reload
    run('nginx -t && systemctl ' + action + ' nginx.service')


def find_files(directory, patterns):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            for pattern in patterns:
                if fnmatch.fnmatch(basename, pattern):
                    yield (root, basename)

def up_project():
    fab_project.rsync_project(
        remote_dir=env.x.remote_path, local_dir='src',
        upload=True, extra_opts='-a ', exclude=('__pycache__'))


def doc(browser=True):
    """ Generate and display document in browser"""
    with fab_context.lcd('doc'):
        local('make')
    if browser:
        with fab_context.lcd('doc'):
            local('cd _build/html && python3 -m http.server 7359 >/dev/null 2>&1 &')
            local('open http://localhost:7359/')


def pg_dump_schema(db=None):
    import getpass
    local_username = getpass.getuser()
    if not db:
        db = env.x.app
    local('pg_dump -s %s  > config/schema.sql' % db)
    local('sed -i s/{local}/{remote}/g config/schema.sql'.format(local=local_username, remote=env.x.app))


def pg_up_schema():
    with _Temp(delete=False, suffix='.dat') as tmp:
        local(
            ("pg_dump --schema-only "
            "--no-owner --no-acl --no-privileges "
            "--file={temp} {db} ")
            .format(temp=tmp.name, db=env.x.app))
        remote_sql = '{path}/tmp/schema.sql'.format(path=env.x.remote_path)
        put(tmp.name, remote_sql)
        run('psql {app} < {sql}'.format(sql=remote_sql, app=env.x.app))


def pg_up_data():
    with _Temp(delete=False, suffix='.dat') as tmp:
        local(
            ("pg_dump --data-only "
            "--no-owner --no-acl --no-privileges "
            "--file={temp} {db} ")
            .format(temp=tmp.name, db=env.x.app))
        remote_sql = '{path}/tmp/data.sql'.format(path=env.x.remote_path)
        put(tmp.name, remote_sql)
        run('psql {app} < {sql}'.format(sql=remote_sql, app=env.x.app))


def pg_setup():
    """
    Create random database and user with password on remote server
    and save credential into production.ini config file
    """
    passwd = rand()
    with open('config/production.ini', 'a') as f:
        lines = [
            '[postgres]',
            'dsn=user={app} dbname={app} password={passwd}'.format(app=env.x.app, passwd=passwd),
            'size=4'
        ]
        for line in lines:
            f.write(line + "\n")
    with_root()
    sql_file = '/tmp/' + rand() + '.sql'
    with _Temp(suffix='.sql') as tmp:
        with open(tmp.name, 'w') as f:
            sql = """
                CREATE USER {app} WITH PASSWORD '{passwd}';
                CREATE DATABASE {app} ENCODING='UTF8';
                GRANT ALL PRIVILEGES ON DATABASE {app} TO {app}
                """.format(passwd=passwd, app=env.x.app)
            f.write(sql)
            f.flush()
            put(tmp.name, sql_file)
    with on_remote('/tmp'):
        run("""su postgres -c 'cat {sql_file} | psql template1' """.format(sql_file=sql_file))
        run('rm ' + sql_file)


def static_link():
    """ Collect static files by symbolic links """
    original_dir = 'src'
    link_dir = 'public'
    for brick_path, basename in find_files(original_dir,[
        '*.css', '*.js', '*.tag'
        '*.png', '*.jpg', '*.gif', '*.json',
        '*.txt', '*.pdf',
    ]):
        static_path = brick_path.replace(original_dir, link_dir)
        if not os.path.exists(static_path):
            os.makedirs(static_path)
        # Make symlinks
        rel_path = os.path.relpath(original_dir, static_path)
        rel = os.path.join(brick_path.replace(original_dir, rel_path), basename)
        dst = os.path.join(static_path, basename)
        print(rel, dst)
        if not os.path.islink(dst):
            os.symlink(rel, dst)
    with fab_context.lcd(link_dir):
        # Remove all broken links
        local('find -xtype l -delete')



from concurrent.futures import ThreadPoolExecutor

from tornado.queues import PriorityQueue, QueueEmpty
from tornado.gen import sleep, coroutine
from tokit import Event, on, logger
from inspect import iscoroutinefunction
from email.mime.text import MIMEText
import smtplib
from email.header import Header

tasks_queue = PriorityQueue()

def put(name, *args, priority=0, **kwargs):
    """
    Schedule a task with given params

    Handlers of event with same name will be used when execute task

    Example::

        @on('task_xyz')
        def do_something(arg1):
            pass

        put('task_xyz', 'val1')
    """
    tasks_queue.put((priority, {'name': name, 'args': args, 'kwargs': kwargs}))


@coroutine
def tasks_consumer(app):
    """
    Check for pending and excute tasks

    A task handler can be coroutine (run in event loop)
    or normal function (run in thread - can be blocking)
    """
    while True:
        yield sleep(0.3)
        try:
            priority, task = tasks_queue.get_nowait()
            handlers = Event.get(task['name']).handlers
            handler = None
            for handler in handlers:
                if iscoroutinefunction(handler):
                    yield handler(
                        app,
                        *task.get('args'),
                        **task.get('kwargs')
                    )
                else:
                    with ThreadPoolExecutor() as executor:
                        yield executor.submit(
                            handler,
                            app,
                            *task.get('args'),
                            **task.get('kwargs')
                        )
            if not handler:
                logger.warn('No handler for task: %s', task['name'])
        except QueueEmpty:
            pass
        else:
            tasks_queue.task_done()


def register_task_runner(app):
    from tornado.ioloop import IOLoop
    IOLoop.current().spawn_callback(lambda: tasks_consumer(app))

@on('send_email')
async def send_email_consumer(app, receipt, body, subject=None):
    if not subject:
        # consider first line as subject
        subject, body = body.split("\n", 1)
    msg = MIMEText(body, 'plain', 'utf-8')
    config = app.config.env['smtp']
    msg['Subject'] = Header(subject, 'utf-8')
    msg['From'] = config['from']
    msg['To'] = receipt
    with smtplib.SMTP(config['host'], config.get('port')) as mailer:
        if config.getboolean('tls'):
            mailer.starttls()
        mailer.login(config['user'], config['password'])
        mailer.send_message(msg)
        mailer.quit()
    logger.debug("Sent email to %s", receipt)


class EmailMixin:

    def send_email(self, template, receipt, **kwargs):
        content = self.render_string(
            template, **kwargs
        ).decode()
        put('send_email', receipt, content)



#!/usr/bin/env python3
import os, sys, re, collections, logging
import time, signal, importlib, inspect, configparser
from contextlib import contextmanager

import tornado.locale
from tornado.httpserver import HTTPServer
import tornado.web
import tornado.websocket
import tornado.netutil
from tornado.ioloop import IOLoop
from tornado import testing

from tokit.utils import Event, AttributeDict, on, to_json

logger = logging.getLogger('tokit')


class Repo:
    """
    Decorator-based registry of objects
    """

    _repo = collections.defaultdict(list)

    def __init__(self, name=None):
        self.name = name

    def __call__(self, obj):
        if not self.name:
            if inspect.isclass(obj):
                parent_cls, *_ = inspect.getmro(obj)
                self.name = parent_cls.__name__
        self._repo[self.name].append(obj)
        return obj

    @classmethod
    def known(cls, name):
        """ Get registered objects
        :return: list
        """
        return cls._repo.get(name, [])


class Registry(type):
    """ A metaclass to make subclasses registry for any class """
    _repo = collections.defaultdict(list)

    def __init__(cls, name, bases, nmspc, **kwarg):
        # Instance object
        super(Registry, cls).__init__(name, bases, nmspc)
        # Register
        repo_name = getattr(cls, 'REPO', None)
        if not repo_name:
            *_, base_cls = bases
            repo_name = base_cls.__name__
        if cls.__name__ == repo_name:
            # Don't add itself, the abstract repo class
            return
        Registry._repo[repo_name].append(cls)

    @classmethod
    def known(mcs, parent_name) -> list:
        """ Get registered subclass of ``parent_name``
        :param str parent_name: pure class name without module prefix
        """
        return Registry._repo[parent_name]


class Request(tornado.web.RequestHandler, metaclass=Registry):
    """
    Base class for handling request
    Class hierarchy is defined right to left, methods are resolved is from left to right
    http://www.ianlewis.org/en/mixins-and-python
    Therefore, mixin if exists, should be added to the left of declare::

        class Foo(BarMixin, Request):
            pass
    """

    URL = None
    """
    Route can be pattern or (pattern, name) or an URLSpec::

        class Post(Request):
            URL = r'/post/.*'
            def get(self, slug):
                pass # Add logic here
    """

    def set_default_headers(self):
        self.set_header('Server', 'Python3')

    def abs_url(self, *args):
        return self.request.protocol + "://" + self.request.host + self.reverse_url(*args)

    def get_template_namespace(self):
        namespace = super(Request, self).get_template_namespace()
        namespace['json'] = to_json
        namespace['url'] = self.abs_url
        return namespace

    def js(self):
        """ List (to preserved ordering) of JS path to to used by layout file """
        return []

    def css(self):
        """ List (to preserved ordering) of CSS path to to used by layout file """
        return []

    def get_request_dict(self, *args):
        """
        Return dict of request arguments, use with standard HTTP form
         """
        return collections.OrderedDict(
            (field, self.get_body_argument(field, None)) for field in args)

    def redirect_referer(self):
        return self.redirect(self.request.headers.get('Referer', '/'))

    @property
    def env(self):
        return self.application.config.env

    @classmethod
    def known(cls):
        """ Get Request's subclasses
        TODO add weight or ordering important routes
        """
        routes = []
        for handler in Registry.known(cls.__name__):
            route = getattr(handler, 'URL', None)
            if not route:
                if not handler.__module__.startswith('_'):
                    logger.debug("Missing route for handler %s.%s",
                                 handler.__module__, handler.__name__)
                continue
            if isinstance(route, str):
                routes.append(tornado.web.URLSpec(route, handler))
            elif isinstance(route, tornado.web.URLSpec):
                routes.append(route)
            else:
                pattern, name = route
                routes.append(tornado.web.URLSpec(pattern, handler, name=name))
        return routes


class Websocket(tornado.websocket.WebSocketHandler, metaclass=Registry):
    def reply(self, _payload=None, **kwargs):
        self.write_message(_payload or to_json(kwargs))

    @property
    def env(self):
        return self.application.config.env


class Module(tornado.web.UIModule, metaclass=Registry):
    """ Subclass this to create a UIModule
    It's available as class name::

        class Table(Module):
            def render(self):
                pass # then in template, you can use {% module Table() %}

    """

    @classmethod
    def known(cls):
        return {c.__name__: c for c in Registry.known(cls.__name__)}


class Static(tornado.web.StaticFileHandler):
    ALLOW_TYPES = (
        'tag', 'js', 'css',
        'png', 'jpg', 'ico', 'svg', 'gif',
        'zip', 'tar', 'tgz', 'txt'
    )
    VALID_PATH = re.compile(r'.*\.({types})$'.format(types='|'.join(ALLOW_TYPES)))

    def validate_absolute_path(self, root, absolute_path):
        """ Add file types checking """
        if not self.VALID_PATH.match(absolute_path):
            raise tornado.web.HTTPError(403, 'Unallowed file type')
        return absolute_path

    @classmethod
    def get_content_version(cls, abspath):
        return super().get_content_version(abspath)[:6]

class Config:
    """ Subclass this to customize runtime config """

    settings = dict(
        compress_response=True,
        static_path='.',
        static_url_prefix='/static/',
        static_handler_class=Static,
        debug=True,
        compiled_template_cache=False,
        cookie_secret='TODO'
    )
    x = AttributeDict()
    root_path = None
    graceful = True
    timezone = 'UTC'
    locale = 'en'

    modules = None
    """ List of dirname, as a module """

    kill_blocking = 2
    """ (second) time to kill process if blocking too long """

    session_timeout = 60 * 24 * 3600
    env_name = None
    env = {}

    def __init__(self, base_file):
        self.root_path = os.path.abspath(os.path.dirname(base_file))
        os.chdir(self.root_path)
        self.settings['static_path'] = os.path.join(self.root_path, self.settings['static_path'])
        lang_path = os.path.join(self.root_path, 'lang')
        if os.path.exists(lang_path):
            tornado.locale.load_translations(lang_path)

    def read_ini(self, cfg_files=None):
        """ Load extra env config
        :param: list cfg_files relative path to project root
        """
        main, *overrides = [os.path.join(self.root_path, '../config', f) for f in cfg_files]
        self.env.read_file(open(main))
        if len(overrides):
            self.env.read(overrides)

    def setup(self):
        self.graceful = self.env['app'].getboolean('graceful')
        self.settings['debug'] = self.env['app'].getboolean('debug')
        log_level = getattr(logging, self.env['app'].get('log_level'))
        logging.basicConfig(level=log_level)
        logger.setLevel(log_level)

        self.settings['cookie_secret'] = self.env['secret'].get('cookie_secret')
        dns_resolver = self.env['app'].get('dns_resolver', 'tornado.netutil.ThreadedResolver')
        tornado.netutil.Resolver.configure(dns_resolver)

        os.environ['TZ'] = self.timezone
        time.tzset()
        tornado.locale.set_default_locale(self.locale)

    def set_env(self, env_name=None):
        self.env = configparser.ConfigParser()
        self.env_name = env_name or os.environ.get('ENV', 'development')
        self.read_ini(['base.ini', self.env_name + '.ini'])
        logging.info('Env: ' + self.env_name)
        Event.get('env').emit(self.env)
        self.setup()
        Event.get('config').emit(self)

    def load_modules(self):
        """
        Import declared modules from config, during this imports, routes
        and other components will be registered
        """
        if not self.modules:
            # Search all modules as a dir,
            _, self.modules, _ = next(os.walk(self.root_path))
            # Exclude special dirnames
            self.modules = [
                m for m in self.modules
                if not (m.startswith('.') or m.startswith('_'))
            ]
        loaded = []
        for m in self.modules:
            try:
                importlib.import_module(m)
                loaded.append(m)
            except TypeError:
                pass
            except SyntaxError as e:
                ex = sys.exc_info()
                logger.error("Broken module: %s %s %s", ex[0].__name__,
                             os.path.basename(
                                 sys.exc_info()[2].tb_frame.f_code.co_filename),
                             ex[2].tb_lineno)
                raise e
        logger.info('Autoloaded modules: %s', loaded)

class App(tornado.web.Application):
    config = None

    @classmethod
    def instance(cls, config):
        config.load_modules()
        Event.get('config').emit(config)
        config.settings['ui_modules'] = Module.known()

        app = App(**config.settings)
        app.config = config
        Event.get('init').emit(app)
        Event.get(config.env_name).emit(app)

        app.add_handlers('.*$', Request.known())
        Event.get('after_init').emit(app)
        return app


def start(host, port, config):
    """
    Entry point for application.
    This setups IOLoop, load classes and run HTTP server
    """

    app = App.instance(config)

    http_server = HTTPServer(app, xheaders=True)
    ioloop = IOLoop.instance()
    http_server.listen(port, host)
    logger.info('Running PID {pid} @ http://{host}:{port}'.format(host=host, pid=os.getpid(), port=port))

    if config.graceful:
        def _graceful():
            """
            Schedule shutdown on next tick
            """
            def _shutdown():
                http_server.stop()
                ioloop.stop()
                logger.info('Stopped')

            ioloop.call_later(1, _shutdown)

        def _on_term(*args):
            ioloop.add_callback_from_signal(_graceful)

        # Automatically kill if anything blocks the process
        signal.signal(signal.SIGTERM, _on_term)
        signal.signal(signal.SIGINT, _on_term)
        ioloop.set_blocking_log_threshold(1)
        ioloop.set_blocking_signal_threshold(config.kill_blocking, action=None)

    try:
        Event.get('start').emit(app)
        ioloop.start()
    except KeyboardInterrupt:
        logger.info('Bye.')
    except Exception as e:
        logger.exception(e)
        logger.info('Server was down.')



import json
import traceback
import functools
import string
import random
import cgitb
import logging


from tornado.websocket import WebSocketHandler
from tornado.gen import coroutine
from tornado.web import HTTPError

from cerberus import Validator

from tokit import Registry, Request, logger, on
from tokit.utils import to_json

SHORT_UUID_RE = '[\-a-zA-Z0-9]{22}'

def parse_json(s):
    try:
        return json.loads(s)
    except ValueError as e:
        raise Exception('Invalid JSON: ' + str(e))


class JsonMixin:
    """
    * Auto parse JSON request body and store in self.data
    * Support render whatever objects as JSON
    """

    def prepare(self):
        self.data = None
        if self.request.body:
            try:
                self.data = parse_json(self.request.body.decode())
            except ValueError as e:
                self.write_exception(e)

    def write_json(self, obj=None, **kwargs):
        if isinstance(obj, list):
            raise ValueError('Lists not accepted for security reasons')
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        ret = kwargs if not obj else obj
        self.write(to_json(ret))


class ErrorMixin:
    """
    Show errors in JSON instead of default HTML

    Always check key "status" from result, will be "ok" or "error"
    """

    def write_error(self, status_code, **kwargs):
        response = {}
        if 'exc_info' in kwargs:
            cls, exception, tb = kwargs['exc_info']
            if hasattr(exception, 'reason'):
                response['reason'] = exception.reason
            if hasattr(exception, 'detail'):
                response['detail'] = exception.detail
            if self.application.config.env.getboolean('app', 'full_trace'):
                print(cgitb.text(kwargs["exc_info"]))
        self.set_status(status_code)
        if isinstance(self, WebSocketHandler):
            self.write_message(response)
        else:
            self.write(response)
            self.finish()


class Resource(ErrorMixin, JsonMixin, Request):
    """
    This must be subclass and mix with a DB Mixin which supports
    `db_query, db_insert` methods

    To add checking, permission by roles
    Overide `def prepare(self)
    """

    REPO = 'Resource'
    URL_PREFIX = '/api'
    TABLE = None

    @coroutine
    def get(self):  # list items
        t, q = self.db_prepare(self.TABLE)
        rows = yield self.db_select(q.fields('*'))
        rows = self.db_serialize(rows)
        self.write_json(length=len(rows), items=rows)

    @coroutine
    def post(self):
        self.set_status(201)
        ret = yield self.db_insert(self.TABLE, self.data)
        self.write_json(ret=ret)


class Item(ErrorMixin, JsonMixin, Request):
    """
    This must be subclass and mix with a DB Mixin which supports
    `db_query, db_update` methods
    """
    REPO = 'Item'
    URL_PREFIX = '/api'
    TABLE = None

    @coroutine
    def get(self, row_id):
        t, q = self.db_prepare(self.TABLE, row_id=row_id)
        rows = yield self.db_select(q.fields('*'))
        self.write_json(item=rows[0])

    @coroutine
    def patch(self, row_id):
        if not self.data:
            raise HTTPError(400, 'No data to patch')
        ret = yield self.db_update(
            self.TABLE, row_id, self.data
        )
        self.write_json(status='ok', detail=ret)

    @coroutine
    def delete(self, row_id):
        ret = yield self.db_delete(self.TABLE, row_id)
        self.write_json(status='ok', detail=ret)


@on('init')
def register_resources_items(self):
    """
    If accept-content is json, run the endpoint
    If accept HTML then return reference documentation
    """
    request_repo = Registry._repo['Request']
    resources = Registry.known('Resource')
    items = Registry.known('Item')

    for resource in resources:
        if not resource.TABLE:
            continue
        if not resource.URL:
            resource.URL = r'^{prefix}/{res}/?$'.format(
                prefix=resource.URL_PREFIX,
                res=resource.TABLE
            )
        request_repo.append(resource)

    logger.debug('Items: %s', [r.TABLE for r in items])
    for item in items:
        if not resource.TABLE:
            continue
        if not item.URL:
            item.URL = r'^{prefix}/{res}/({id})/?$'.format(
                prefix=resource.URL_PREFIX,
                res=resource.TABLE,
                id=SHORT_UUID_RE)
        request_repo.append(item)


class PlainApi(ErrorMixin, JsonMixin, Request):
    REPO = 'Request'
    TABLE = None


class ValidationError(HTTPError):

    def __init__(self, errors, *args, **kwargs):
        self.status_code = 400
        self.log_message = None
        self.reason = 'Validation failed {}'.format(', '.join(tuple(errors.keys())))
        self.detail = errors


class ValidatorMixin:

    SCHEMA = None

    def validate(self):
        if not self.SCHEMA:
            raise ValueError("Cannot validate because schema isn't set")
        self.validator = Validator(self.SCHEMA)
        if not self.validator.validate(self.data):
            raise ValidationError(self.validator.errors)


class CreateMixin(ValidatorMixin):
    """
    Validate and create record
    from untrusted sources
    """

    @coroutine
    def on_validated(self, data):
        inserted_id = yield self.db_insert(self.TABLE, **data)
        return inserted_id

    @coroutine
    def post(self):
        self.validate()
        try:
            ret = yield self.on_validated(self.validator.document)
        except self.DbIntegrityError as e:
            self.set_status(400)
            self.write_json({'reason': str(e)})
        except self.DbError as e:
            self.set_status(500)
            self.write_json({'reason': str(e)})
        else:
            self.set_status(201)
            self.write_json({
                'id': ret
            })

class UpdateMixin(ValidatorMixin):
    """
    Validate and update record
    from untrusted sources
    """

    @coroutine
    def on_validated(self, data, uid):
        changes = yield self.db_update(self.TABLE, uid, data)
        return changes

    @coroutine
    def put(self, uid):
        self.validate()
        try:
            changes = yield self.on_validated(self.validator.document, uid)
        except self.DbError as e:
            self.set_status(500)
            self.write_json({'reason': str(e)})
        else:
            self.set_status(200)
            self.finish()



from setuptools import setup

# Thanks to http://peterdowns.com/posts/first-time-with-pypi.html
tokit_version = '0.6.4'

setup(
    name='tokit',
    packages=['tokit'],
    version=tokit_version,
    description='A kit for development with Tornado web framework',
    author='Giang Manh',
    author_email='manhgd@yahoo.com',
    url='https://github.com/manhg/tokit',
    download_url='https://github.com/manhg/tokit/tarball/' + tokit_version,
    keywords=['tornado', 'web', 'tokit'],
    install_requires = [
        "tornado>=4.3",
        "shortuuid"
    ],
    package_data={'tokit': ['js/*.js', 'skeleton/*']},
    classifiers=[
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ]
)
